@@ -1,5 +1,5 @@
 /* dd -- convert a file while copying it.
-   Copyright (C) 85, 90, 91, 1995-2000 Free Software Foundation, Inc.
+   Copyright (C) 85, 90, 91, 1995-2001 Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -48,15 +48,23 @@
 # define SIGINFO SIGUSR1
 #endif
 
+#ifndef S_TYPEISSHM
+# define S_TYPEISSHM(Stat_ptr) 0
+#endif
+
 #define ROUND_UP_OFFSET(X, M) ((M) - 1 - (((X) + (M) - 1) % (M)))
 #define PTR_ALIGN(Ptr, M) ((Ptr) \
 			   + ROUND_UP_OFFSET ((char *)(Ptr) - (char *)0, (M)))
 
 #define max(a, b) ((a) > (b) ? (a) : (b))
-#define output_char(c) \
-  do { \
-  obuf[oc++] = (c); if (oc >= output_blocksize) write_output (); \
-  } while (0)
+#define output_char(c)				\
+  do						\
+    {						\
+      obuf[oc++] = (c);				\
+      if (oc >= output_blocksize)		\
+	write_output ();			\
+    }						\
+  while (0)
 
 /* Default input and output blocksize. */
 #define DEFAULT_BLOCKSIZE 512
@@ -100,7 +108,7 @@ static size_t conversion_blocksize = 0;
 static uintmax_t skip_records = 0;
 
 /* Skip this many records of `output_blocksize' bytes before output. */
-static uintmax_t seek_record = 0;
+static uintmax_t seek_records = 0;
 
 /* Copy only this many records.  The default is effectively infinity.  */
 static uintmax_t max_records = (uintmax_t) -1;
@@ -298,7 +306,7 @@ Copy a file, converting and formatting according to the options.\n\
       --help      display this help and exit\n\
       --version   output version information and exit\n\
 \n\
-BYTES may be followed by the following multiplicative suffixes:\n\
+BLOCKS and BYTES may be followed by the following multiplicative suffixes:\n\
 xM M, c 1, w 2, b 512, kD 1000, k 1024, MD 1,000,000, M 1,048,576,\n\
 GD 1,000,000,000, G 1,073,741,824, and so on for T, P, E, Z, Y.\n\
 Each KEYWORD may be:\n\
@@ -583,7 +591,7 @@ scanargs (int argc, char **argv)
 	  else if (STREQ (name, "skip"))
 	    skip_records = n;
 	  else if (STREQ (name, "seek"))
-	    seek_record = n;
+	    seek_records = n;
 	  else if (STREQ (name, "count"))
 	    max_records = n;
 	  else
@@ -734,7 +742,7 @@ buggy_lseek_support (int fdesc)
      on some special files but doesn't return an error, either.
      In particular, the Linux tape drivers are a problem.
      For example, when I did the following using dd-4.0y or earlier on a
-     Linux-2.2.17 system with a Exabyte SCSI tape drive:
+     Linux-2.2.17 system with an Exabyte SCSI tape drive:
 
        dev=/dev/nst0
        reset='mt -f $dev rewind; mt -f $dev fsf 1'
@@ -758,22 +766,20 @@ static void
 skip (int fdesc, char *file, uintmax_t records, size_t blocksize,
       unsigned char *buf)
 {
-  off_t o;
+  off_t offset = records * blocksize;
 
   /* Try lseek and if an error indicates it was an inappropriate
      operation, fall back on using read.  Some broken versions of
      lseek may return zero, so count that as an error too as a valid
      zero return is not possible here.  */
-  o = records * blocksize;
-  if (o / blocksize != records
-      || buggy_lseek_support (fdesc)
-      || lseek (fdesc, o, SEEK_CUR) <= 0)
-    {
-      while (records-- > 0)
-	{
-	  int nread;
 
-	  nread = safe_read (fdesc, buf, blocksize);
+  if (offset / blocksize != records
+      || buggy_lseek_support (fdesc)
+      || lseek (fdesc, offset, SEEK_CUR) <= 0)
+    {
+      while (records--)
+	{
+	  ssize_t nread = safe_read (fdesc, buf, blocksize);
 	  if (nread < 0)
 	    {
 	      error (0, errno, _("reading %s"), quote (file));
@@ -791,10 +797,10 @@ skip (int fdesc, char *file, uintmax_t records, size_t blocksize,
 /* Copy NREAD bytes of BUF, with no conversions.  */
 
 static void
-copy_simple (unsigned char *buf, int nread)
+copy_simple (unsigned char const *buf, int nread)
 {
   int nfree;			/* Number of unused bytes in `obuf'.  */
-  unsigned char *start = buf; /* First uncopied char in BUF.  */
+  const unsigned char *start = buf; /* First uncopied char in BUF.  */
 
   do
     {
@@ -818,7 +824,7 @@ copy_simple (unsigned char *buf, int nread)
    replacing the newline with trailing spaces).  */
 
 static void
-copy_with_block (unsigned char *buf, size_t nread)
+copy_with_block (unsigned char const *buf, size_t nread)
 {
   size_t i;
 
@@ -850,7 +856,7 @@ copy_with_block (unsigned char *buf, size_t nread)
    with a newline).  */
 
 static void
-copy_with_unblock (unsigned char *buf, size_t nread)
+copy_with_unblock (unsigned char const *buf, size_t nread)
 {
   size_t i;
   unsigned char c;
@@ -890,7 +896,7 @@ dd_copy (void)
   unsigned char *ibuf, *bufstart; /* Input buffer. */
   unsigned char *real_buf;	  /* real buffer address before alignment */
   unsigned char *real_obuf;
-  int nread;			/* Bytes read in the current block. */
+  ssize_t nread;		/* Bytes read in the current block. */
   int exit_status = 0;
   size_t page_size = getpagesize ();
   size_t n_bytes_read;
@@ -935,7 +941,7 @@ dd_copy (void)
   if (skip_records != 0)
     skip (STDIN_FILENO, input_file, skip_records, input_blocksize, ibuf);
 
-  if (seek_record != 0)
+  if (seek_records != 0)
     {
       /* FIXME: this loses for
 	 % ./dd if=dd seek=1 |:
@@ -944,8 +950,7 @@ dd_copy (void)
 	 0+0 records out
 	 */
 
-	 skip (STDOUT_FILENO, output_file, seek_record, output_blocksize,
-	       obuf);
+      skip (STDOUT_FILENO, output_file, seek_records, output_blocksize, obuf);
     }
 
   if (max_records == 0)
@@ -990,7 +995,7 @@ dd_copy (void)
 	    }
 	}
 
-      n_bytes_read = (size_t) nread;
+      n_bytes_read = nread;
 
       if (n_bytes_read < input_blocksize)
 	{
@@ -1110,7 +1115,7 @@ main (int argc, char **argv)
   /* Arrange to close stdout if parse_long_options exits.  */
   atexit (close_stdout_wrapper);
 
-  parse_long_options (argc, argv, PROGRAM_NAME, GNU_PACKAGE, VERSION,
+  parse_long_options (argc, argv, PROGRAM_NAME, PACKAGE, VERSION,
 		      AUTHORS, usage);
 
   /* Don't close stdout on exit from here on.  */
@@ -1138,27 +1143,41 @@ main (int argc, char **argv)
       mode_t perms = S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH | S_IWOTH;
       int opts
 	= (O_CREAT
-	   | (seek_record || (conversions_mask & C_NOTRUNC) ? 0 : O_TRUNC));
+	   | (seek_records || (conversions_mask & C_NOTRUNC) ? 0 : O_TRUNC));
 
       /* Open the output file with *read* access only if we might
 	 need to read to satisfy a `seek=' request.  If we can't read
 	 the file, go ahead with write-only access; it might work.  */
-      if ((! seek_record
+      if ((! seek_records
 	   || open_fd (STDOUT_FILENO, output_file, O_RDWR | opts, perms) < 0)
 	  && open_fd (STDOUT_FILENO, output_file, O_WRONLY | opts, perms) < 0)
 	error (1, errno, _("opening %s"), quote (output_file));
+
 #if HAVE_FTRUNCATE
-      if (seek_record != 0 && !(conversions_mask & C_NOTRUNC))
+      if (seek_records != 0 && !(conversions_mask & C_NOTRUNC))
 	{
-	  off_t o = seek_record * output_blocksize;
-	  if (o / output_blocksize != seek_record)
+	  struct stat stdout_stat;
+	  off_t o = seek_records * output_blocksize;
+	  if (o / output_blocksize != seek_records)
 	    error (1, 0, _("file offset out of range"));
-	  if (ftruncate (STDOUT_FILENO, o) < 0)
+
+	  if (fstat (STDOUT_FILENO, &stdout_stat) != 0)
+	    error (1, errno, _("cannot fstat %s"), quote (output_file));
+
+	  /* Complain only when ftruncate fails on a regular file, a
+	     directory, or a shared memory object, as the 2000-08
+	     POSIX draft specifies ftruncate's behavior only for these
+	     file types.  For example, do not complain when Linux 2.4
+	     ftruncate fails on /dev/fd0.  */
+	  if (ftruncate (STDOUT_FILENO, o) != 0
+	      && (S_ISREG (stdout_stat.st_mode)
+		  || S_ISDIR (stdout_stat.st_mode)
+		  || S_TYPEISSHM (&stdout_stat)))
 	    {
 	      char buf[LONGEST_HUMAN_READABLE + 1];
-	      error (1, errno, _("advancing past %s blocks in output file %s"),
-				 human_readable (seek_record, buf, 1, 1),
-				 quote (output_file));
+	      error (1, errno, _("advancing past %s bytes in output file %s"),
+		     human_readable (o, buf, 1, 1),
+		     quote (output_file));
 	    }
 	}
 #endif
