@@ -1,5 +1,5 @@
 /* `dir', `vdir' and `ls' directory listing programs for GNU.
-   Copyright (C) 85, 88, 90, 91, 1995-2000 Free Software Foundation, Inc.
+   Copyright (C) 85, 88, 90, 91, 1995-2001 Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -93,12 +93,6 @@
 # define iswprint(wc) 1
 #endif
 
-/* Some systems, like BeOS, have multibyte encodings but lack mbstate_t.  */
-#if HAVE_MBRTOWC && defined mbstate_t
-# define mbrtowc(pwc, s, n, ps) (mbrtowc) (pwc, s, n, 0)
-# define mbsinit(ps) 1
-#endif
-
 #ifndef HAVE_DECL_WCWIDTH
 "this configure-time declaration test was not run"
 #endif
@@ -197,12 +191,32 @@ int rpl_lstat PARAMS((const char *, struct stat *));
 # define lstat(Name, Stat_buf) rpl_lstat(Name, Stat_buf)
 #endif
 
+#if defined _DIRENT_HAVE_D_TYPE || defined DTTOIF
+# define HAVE_STRUCT_DIRENT_D_TYPE 1
+# define DT_INIT(Val) = Val
+#else
+# define HAVE_STRUCT_DIRENT_D_TYPE 0
+# define DT_INIT(Val) /* empty */
+#endif
+
+#ifdef ST_MTIM_NSEC
+# define TIMESPEC_NS(timespec) ((timespec).ST_MTIM_NSEC)
+#else
+# define TIMESPEC_NS(timespec) 0
+#endif
+
 enum filetype
   {
-    symbolic_link,
-    directory,
-    arg_directory,		/* Directory given as command line arg. */
-    normal			/* All others. */
+    unknown DT_INIT (DT_UNKNOWN),
+    fifo DT_INIT (DT_FIFO),
+    chardev DT_INIT (DT_CHR),
+    directory DT_INIT (DT_DIR),
+    blockdev DT_INIT (DT_BLK),
+    normal DT_INIT (DT_REG),
+    symbolic_link DT_INIT (DT_LNK),
+    sock DT_INIT (DT_SOCK),
+    arg_directory DT_INIT (2 * (DT_UNKNOWN | DT_FIFO | DT_CHR | DT_DIR | DT_BLK
+				| DT_REG | DT_LNK | DT_SOCK))
   };
 
 struct fileinfo
@@ -290,8 +304,8 @@ static int rev_cmp_version PARAMS ((const struct fileinfo *file2,
 				    const struct fileinfo *file1));
 static int decode_switches PARAMS ((int argc, char **argv));
 static int file_interesting PARAMS ((const struct dirent *next));
-static uintmax_t gobble_file PARAMS ((const char *name, int explicit_arg,
-				      const char *dirname));
+static uintmax_t gobble_file PARAMS ((const char *name, enum filetype type,
+				      int explicit_arg, const char *dirname));
 static void print_color_indicator PARAMS ((const char *name, unsigned int mode,
 					   int linkok));
 static void put_indicator PARAMS ((const struct bin_str *ind));
@@ -369,10 +383,11 @@ struct pending
 
 static struct pending *pending_dirs;
 
-/* Current time (seconds since 1970).  When we are printing a file's time,
-   include the year if it is more than 6 months before this time.  */
+/* Current time in seconds and nanoseconds since 1970, updated as
+   needed when deciding whether a file is recent.  */
 
-static time_t current_time;
+static time_t current_time = TYPE_MINIMUM (time_t);
+static int current_time_ns = -1;
 
 /* The number of digits to use for block sizes.
    4, or more if needed for bigger numbers.  */
@@ -629,17 +644,43 @@ static int line_length;
 
 static int format_needs_stat;
 
+/* Similar to `format_needs_stat', but set if only the file type is
+   needed.  */
+
+static int format_needs_type;
+
+/* strftime formats for non-recent and recent files, respectively, in
+   -l output.  */
+
+static char const *long_time_format[2];
+
 /* The exit status to use if we don't get any fatal errors. */
 
 static int exit_status;
 
+/* For long options that have no equivalent short option, use a
+   non-character as a pseudo short option, starting with CHAR_MAX + 1.  */
+enum
+{
+  BLOCK_SIZE_OPTION = CHAR_MAX + 1,
+  COLOR_OPTION,
+  FORMAT_OPTION,
+  FULL_TIME_OPTION,
+  INDICATOR_STYLE_OPTION,
+  QUOTING_STYLE_OPTION,
+  SHOW_CONTROL_CHARS_OPTION,
+  SI_OPTION,
+  SORT_OPTION,
+  TIME_OPTION
+};
+
 static struct option const long_options[] =
 {
   {"all", no_argument, 0, 'a'},
   {"escape", no_argument, 0, 'b'},
   {"directory", no_argument, 0, 'd'},
   {"dired", no_argument, 0, 'D'},
-  {"full-time", no_argument, &full_time, 1},
+  {"full-time", no_argument, 0, FULL_TIME_OPTION},
   {"human-readable", no_argument, 0, 'h'},
   {"inode", no_argument, 0, 'i'},
   {"kilobytes", no_argument, 0, 'k'},
@@ -653,21 +694,21 @@ static struct option const long_options[] =
   {"ignore-backups", no_argument, 0, 'B'},
   {"classify", no_argument, 0, 'F'},
   {"file-type", no_argument, 0, 'p'},
-  {"si", no_argument, 0, 'H'},
+  {"si", no_argument, 0, SI_OPTION},
   {"ignore", required_argument, 0, 'I'},
-  {"indicator-style", required_argument, 0, 14},
+  {"indicator-style", required_argument, 0, INDICATOR_STYLE_OPTION},
   {"dereference", no_argument, 0, 'L'},
   {"literal", no_argument, 0, 'N'},
   {"quote-name", no_argument, 0, 'Q'},
-  {"quoting-style", required_argument, 0, 15},
+  {"quoting-style", required_argument, 0, QUOTING_STYLE_OPTION},
   {"recursive", no_argument, 0, 'R'},
-  {"format", required_argument, 0, 12},
-  {"show-control-chars", no_argument, 0, 16},
-  {"sort", required_argument, 0, 10},
+  {"format", required_argument, 0, FORMAT_OPTION},
+  {"show-control-chars", no_argument, 0, SHOW_CONTROL_CHARS_OPTION},
+  {"sort", required_argument, 0, SORT_OPTION},
   {"tabsize", required_argument, 0, 'T'},
-  {"time", required_argument, 0, 11},
-  {"color", optional_argument, 0, 13},
-  {"block-size", required_argument, 0, 17},
+  {"time", required_argument, 0, TIME_OPTION},
+  {"color", optional_argument, 0, COLOR_OPTION},
+  {"block-size", required_argument, 0, BLOCK_SIZE_OPTION},
   {GETOPT_HELP_OPTION_DECL},
   {GETOPT_VERSION_OPTION_DECL},
   {NULL, 0, NULL, 0}
@@ -828,7 +869,6 @@ main (int argc, char **argv)
   dir_defaulted = 1;
   print_dir_name = 1;
   pending_dirs = 0;
-  current_time = time ((time_t *) 0);
 
   i = decode_switches (argc, argv);
 
@@ -845,8 +885,9 @@ main (int argc, char **argv)
 
   format_needs_stat = sort_type == sort_time || sort_type == sort_size
     || format == long_format
-    || trace_links || trace_dirs || indicator_style != none
-    || print_block_size || print_inode || print_with_color;
+    || trace_links || trace_dirs || print_block_size || print_inode;
+  format_needs_type = (format_needs_stat == 0
+		       && (print_with_color || indicator_style != none));
 
   if (dired && format == long_format)
     {
@@ -864,13 +905,13 @@ main (int argc, char **argv)
     dir_defaulted = 0;
   for (; i < argc; i++)
     {
-      gobble_file (argv[i], 1, "");
+      gobble_file (argv[i], unknown, 1, "");
     }
 
   if (dir_defaulted)
     {
       if (immediate_dirs)
-	gobble_file (".", 1, "");
+	gobble_file (".", directory, 1, "");
       else
 	queue_directory (".", 0);
     }
@@ -1087,6 +1128,12 @@ decode_switches (int argc, char **argv)
 	  break;
 
 	case 'H':
+	  error (0, 0,
+		 _("\
+Warning: the meaning of `-H' will change in the future to conform to POSIX.\n\
+Use `--si' for the old meaning."));
+	  /* Fall through.  */
+	case SI_OPTION:
 	  output_block_size = -1000;
 	  break;
 
@@ -1230,20 +1277,25 @@ decode_switches (int argc, char **argv)
 	  format = one_per_line;
 	  break;
 
-	case 10:		/* --sort */
+	case SORT_OPTION:
 	  sort_type = XARGMATCH ("--sort", optarg, sort_args, sort_types);
 	  sort_type_specified = 1;
 	  break;
 
-	case 11:		/* --time */
+	case TIME_OPTION:
 	  time_type = XARGMATCH ("--time", optarg, time_args, time_types);
 	  break;
 
-	case 12:		/* --format */
+	case FORMAT_OPTION:
 	  format = XARGMATCH ("--format", optarg, format_args, format_types);
 	  break;
 
-	case 13:		/* --color */
+	case FULL_TIME_OPTION:
+	  format = long_format;
+	  full_time = 1;
+	  break;
+
+	case COLOR_OPTION:
 	  if (optarg)
 	    i = XARGMATCH ("--color", optarg, color_args, color_types);
 	  else
@@ -1264,24 +1316,24 @@ decode_switches (int argc, char **argv)
 	    }
 	  break;
 
-	case 14:		/* --indicator-style */
+	case INDICATOR_STYLE_OPTION:
 	  indicator_style = XARGMATCH ("--indicator-style", optarg,
 				       indicator_style_args,
 				       indicator_style_types);
 	  break;
 
-	case 15:		/* --quoting-style */
+	case QUOTING_STYLE_OPTION:
 	  set_quoting_style (NULL,
 			     XARGMATCH ("--quoting-style", optarg,
 					quoting_style_args,
 					quoting_style_vals));
 	  break;
 
-	case 16:
+	case SHOW_CONTROL_CHARS_OPTION:
 	  qmark_funny_chars = 0;
 	  break;
 
-	case 17:
+	case BLOCK_SIZE_OPTION:
 	  human_block_size (optarg, 1, &output_block_size);
 	  break;
 
@@ -1318,6 +1370,18 @@ decode_switches (int argc, char **argv)
       sort_type = sort_time;
     }
 
+  if (format == long_format)
+    {
+      if (full_time)
+	long_time_format[0] = long_time_format[1] =
+	  dcgettext (NULL, "%a %b %d %H:%M:%S %Y", LC_TIME);
+      else
+	{
+	  long_time_format[0] = dcgettext (NULL, "%b %e  %Y", LC_TIME);
+	  long_time_format[1] = dcgettext (NULL, "%b %e %H:%M", LC_TIME);
+	}
+    }
+
   return optind;
 }
 
@@ -1701,7 +1765,17 @@ print_dir (const char *name, const char *realname)
 
   while ((next = readdir (reading)) != NULL)
     if (file_interesting (next))
-      total_blocks += gobble_file (next->d_name, 0, name);
+      {
+	enum filetype type = unknown;
+
+#if HAVE_STRUCT_DIRENT_D_TYPE
+	if (next->d_type == DT_DIR || next->d_type == DT_CHR
+	    || next->d_type == DT_BLK || next->d_type == DT_SOCK
+	    || next->d_type == DT_FIFO)
+	  type = next->d_type;
+#endif
+	total_blocks += gobble_file (next->d_name, type, 0, name);
+      }
 
   if (CLOSEDIR (reading))
     {
@@ -1812,10 +1886,10 @@ clear_files (void)
    Return the number of blocks that the file occupies.  */
 
 static uintmax_t
-gobble_file (const char *name, int explicit_arg, const char *dirname)
+gobble_file (const char *name, enum filetype type, int explicit_arg,
+	     const char *dirname)
 {
   register uintmax_t blocks;
-  register int val;
   register char *path;
 
   if (files_index == nfiles)
@@ -1829,9 +1903,15 @@ gobble_file (const char *name, int explicit_arg, const char *dirname)
   files[files_index].linkmode = 0;
   files[files_index].linkok = 0;
 
-  if (explicit_arg || format_needs_stat)
+  /* FIXME: this use of ls: `mkdir a; touch a/{b,c,d}; ls -R a'
+     shouldn't require that ls stat b, c, and d -- at least
+     not on systems with usable d_type.  The problem is that
+     format_needs_stat is set, because of the -R.  */
+  if (explicit_arg || format_needs_stat
+      || (format_needs_type && type == unknown))
     {
       /* `path' is the absolute pathname of this file. */
+      int val;
 
       if (name[0] == '/' || dirname[0] == 0)
 	path = (char *) name;
@@ -1841,23 +1921,9 @@ gobble_file (const char *name, int explicit_arg, const char *dirname)
 	  attach (path, dirname, name);
 	}
 
-      if (trace_links)
-	{
-	  val = stat (path, &files[files_index].stat);
-	  if (val < 0)
-	    {
-	      /* Perhaps a symbolically-linked to file doesn't exist; stat
-		 the link instead. */
-	      val = lstat (path, &files[files_index].stat);
-	    }
-	}
-      else
-	{
-	  val = lstat (path, &files[files_index].stat);
-#if USE_ACL
-	  files[files_index].have_acl = (acl (path, GETACLCNT, 0, NULL) > 4);
-#endif
-	}
+      val = (trace_links
+	     ? stat (path, &files[files_index].stat)
+	     : lstat (path, &files[files_index].stat));
 
       if (val < 0)
 	{
@@ -1866,6 +1932,13 @@ gobble_file (const char *name, int explicit_arg, const char *dirname)
 	  return 0;
 	}
 
+#if USE_ACL
+      if (format == long_format)
+	files[files_index].have_acl =
+	  (! S_ISLNK (files[files_index].stat.st_mode)
+	   && 4 < acl (path, GETACLCNT, 0, NULL));
+#endif
+
       if (S_ISLNK (files[files_index].stat.st_mode)
 	  && (explicit_arg || format == long_format || check_symlink_color))
 	{
@@ -1937,7 +2010,13 @@ gobble_file (const char *name, int explicit_arg, const char *dirname)
       }
     }
   else
-    blocks = 0;
+    {
+      files[files_index].filetype = type;
+#if HAVE_STRUCT_DIRENT_D_TYPE
+      files[files_index].stat.st_mode = DTTOIF (type);
+#endif
+      blocks = 0;
+    }
 
   files[files_index].name = xstrdup (name);
   files_index++;
@@ -2024,9 +2103,7 @@ static void
 extract_dirs_from_files (const char *dirname, int recursive)
 {
   register int i, j;
-  int dirlen;
 
-  dirlen = strlen (dirname) + 2;
   /* Queue the directories last one first, because queueing reverses the
      order.  */
   for (i = files_index - 1; i >= 0; i--)
@@ -2284,6 +2361,80 @@ print_current_files (void)
     }
 }
 
+/* Return the expected number of columns in a long-format time stamp,
+   or zero if it cannot be calculated.  */
+
+static int
+long_time_expected_width (void)
+{
+  static int width = -1;
+
+  if (width < 0)
+    {
+      time_t epoch = 0;
+      struct tm const *tm = localtime (&epoch);
+      char const *fmt = long_time_format[0];
+      char initbuf[100];
+      char *buf = initbuf;
+      size_t bufsize = sizeof initbuf;
+      size_t len;
+
+      for (;;)
+	{
+	  *buf = '\1';
+	  len = strftime (buf, bufsize, fmt, tm);
+	  if (len || ! *buf)
+	    break;
+	  buf = alloca (bufsize *= 2);
+	}
+
+      width = mbsnwidth (buf, len, 0);
+      if (width < 0)
+	width = 0;
+    }
+
+  return width;
+}
+
+/* Get the current time.  */
+
+static void
+get_current_time (void)
+{
+#if HAVE_CLOCK_GETTIME && defined CLOCK_REALTIME
+  {
+    struct timespec timespec;
+    if (clock_gettime (CLOCK_REALTIME, &timespec) == 0)
+      {
+	current_time = timespec.tv_sec;
+	current_time_ns = timespec.tv_nsec;
+	return;
+      }
+  }
+#endif
+
+  /* The clock does not have nanosecond resolution, so get the maximum
+     possible value for the current time that is consistent with the
+     reported clock.  That way, files are not considered to be in the
+     future merely because their time stamps have higher resolution
+     than the clock resolution.  */
+
+#if HAVE_GETTIMEOFDAY
+  {
+    struct timeval timeval;
+    if (gettimeofday (&timeval, NULL) == 0)
+      {
+	current_time = timeval.tv_sec;
+	current_time_ns = timeval.tv_usec * 1000 + 999;
+	return;
+      }
+  }
+#endif
+
+  current_time = time (NULL);
+  current_time_ns = 999999999;
+}
+
 static void
 print_long_format (const struct fileinfo *f)
 {
@@ -2303,8 +2454,8 @@ print_long_format (const struct fileinfo *f)
   size_t s;
   char *p;
   time_t when;
+  int when_ns IF_LINT (= 0);
   struct tm *when_local;
-  const char *fmt;
   char *user_name;
 
 #if HAVE_ST_DM_MODE
@@ -2321,38 +2472,18 @@ print_long_format (const struct fileinfo *f)
     {
     case time_ctime:
       when = f->stat.st_ctime;
+      when_ns = TIMESPEC_NS (f->stat.st_ctim);
       break;
     case time_mtime:
       when = f->stat.st_mtime;
+      when_ns = TIMESPEC_NS (f->stat.st_mtim);
       break;
     case time_atime:
       when = f->stat.st_atime;
+      when_ns = TIMESPEC_NS (f->stat.st_atim);
       break;
     }
 
-  if (full_time)
-    {
-      fmt = "%a %b %d %H:%M:%S %Y";
-    }
-  else
-    {
-      if (current_time > when + 6L * 30L * 24L * 60L * 60L	/* Old. */
-	  || current_time < when - 60L * 60L)	/* In the future. */
-	{
-	  /* The file is fairly old or in the future.
-	     POSIX says the cutoff is 6 months old;
-	     approximate this by 6*30 days.
-	     Allow a 1 hour slop factor for what is considered "the future",
-	     to allow for NFS server/client clock disagreement.
-	     Show the year instead of the time of day.  */
-	  fmt = "%b %e  %Y";
-	}
-      else
-	{
-	  fmt = "%b %e %H:%M";
-	}
-    }
-
   p = buf;
 
   if (print_inode)
@@ -2408,14 +2539,37 @@ print_long_format (const struct fileinfo *f)
 
   p += strlen (p);
 
-  /* Use strftime rather than ctime, because the former can produce
-     locale-dependent names for the weekday (%a) and month (%b).  */
-
   if ((when_local = localtime (&when)))
     {
-      while (! (s = strftime (p, buf + bufsize - p - 1, fmt, when_local)))
+      time_t six_months_ago;
+      int recent;
+      char const *fmt;
+
+      /* If the file appears to be in the future, update the current
+	 time, in case the file happens to have been modified since
+	 the last time we checked the clock.  */
+      if (current_time < when
+	  || (current_time == when && current_time_ns < when_ns))
+	get_current_time ();
+
+      /* Consider a time to be recent if it is within the past six
+	 months.  A Gregorian year has 365.2425 * 24 * 60 * 60 ==
+	 31556952 seconds on the average.  Write this value as an
+	 integer constant to avoid floating point hassles.  */
+      six_months_ago = current_time - 31556952 / 2;
+      recent = (six_months_ago <= when
+		&& (when < current_time
+		    || (when == current_time && when_ns <= current_time_ns)));
+      fmt = long_time_format[recent];
+
+      for (;;)
 	{
-	  char *newbuf = (char *) alloca (bufsize *= 2);
+	  char *newbuf;
+	  *p = '\1';
+	  s = strftime (p, buf + bufsize - p - 1, fmt, when_local);
+	  if (s || ! *p)
+	    break;
+	  newbuf = alloca (bufsize *= 2);
 	  memcpy (newbuf, buf, p - buf);
 	  p = newbuf + (p - buf);
 	  buf = newbuf;
@@ -2432,7 +2586,7 @@ print_long_format (const struct fileinfo *f)
       /* The time cannot be represented as a local time;
 	 print it as a huge integer number of seconds.  */
       char hbuf[LONGEST_HUMAN_READABLE + 1];
-      int width = full_time ? 24 : 12;
+      int width = long_time_expected_width ();
 
       if (when < 0)
 	{
@@ -3109,7 +3263,7 @@ Sort entries alphabetically if none of -cftuSUX nor --sort.\n\
       --block-size=SIZE      use SIZE-byte blocks\n\
   -B, --ignore-backups       do not list implied entries ending with ~\n\
   -c                         with -lt: sort by, and show, ctime (time of last\n\
-				 modification of file status information)\n\
+                               modification of file status information)\n\
                                with -l: show ctime and sort by name\n\
                                otherwise: sort by ctime\n\
   -C                         list entries by columns\n\
@@ -3127,7 +3281,9 @@ Sort entries alphabetically if none of -cftuSUX nor --sort.\n\
   -g                         (ignored)\n\
   -G, --no-group             inhibit display of group information\n\
   -h, --human-readable  print sizes in human readable format (e.g., 1K 234M 2G)\n\
-  -H, --si                   likewise, but use powers of 1000 not 1024\n\
+      --si                   likewise, but use powers of 1000 not 1024\n\
+  -H                         same as `--si' for now; soon to change\n\
+                               to conform to POSIX\n\
       --indicator-style=WORD append indicator with style WORD to entry names:\n\
                                none (default), classify (-F), file-type (-p)\n\
   -i, --inode                print index number of each file\n\
