@@ -1,5 +1,5 @@
 /* `rm' file deletion utility for GNU.
-   Copyright (C) 88, 90, 91, 94, 95, 96, 97, 1998 Free Software Foundation, Inc.
+   Copyright (C) 88, 90, 91, 1994-2000 Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -22,9 +22,9 @@
 
    In the `usual' case, RM saves no state for directories it is processing.
    When a removal fails (either due to an error or to an interactive `no'
-   reply), the failure is noted (see description of `ht' in remove_cwd_entries)
-   so that when/if the containing directory is reopened, RM doesn't try to
-   remove the entry again.
+   reply), the failure is noted (see description of `ht' in remove.c's
+   remove_cwd_entries function) so that when/if the containing directory
+   is reopened, RM doesn't try to remove the entry again.
 
    RM may delete arbitrarily deep hierarchies -- even ones in which file
    names (from root to leaf) are longer than the system-imposed maximum.
@@ -32,7 +32,7 @@
    removing the entries in that directory.
 
    RM detects directory cycles by maintaining a table of the currently
-   active directories.  See the description of active_dir_map below.
+   active directories.  See the description of active_dir_map in remove.c.
 
    RM is careful to avoid forming full file names whenever possible.
    A full file name is formed only when it is about to be used -- e.g.
@@ -49,23 +49,22 @@
 #include <sys/types.h>
 #include <assert.h>
 
-#include "save-cwd.h"
 #include "system.h"
-#include "closeout.h"
 #include "error.h"
 #include "remove.h"
+#include "save-cwd.h"
+
+/* The official name of this program (e.g., no `g' prefix).  */
+#define PROGRAM_NAME "rm"
+
+#define AUTHORS \
+  "Paul Rubin, David MacKenzie, Richard Stallman, and Jim Meyering"
 
 void strip_trailing_slashes ();
 
 /* Name this program was run with.  */
 char *program_name;
 
-/* If nonzero, display usage information and exit.  */
-static int show_help;
-
-/* If nonzero, print the version on standard output and exit.  */
-static int show_version;
-
 static struct option const long_opts[] =
 {
   {"directory", no_argument, NULL, 'd'},
@@ -73,12 +72,12 @@ static struct option const long_opts[] =
   {"interactive", no_argument, NULL, 'i'},
   {"recursive", no_argument, NULL, 'r'},
   {"verbose", no_argument, NULL, 'v'},
-  {"help", no_argument, &show_help, 1},
-  {"version", no_argument, &show_version, 1},
+  {GETOPT_HELP_OPTION_DECL},
+  {GETOPT_VERSION_OPTION_DECL},
   {NULL, 0, NULL, 0}
 };
 
-static void
+void
 usage (int status)
 {
   if (status != 0)
@@ -97,9 +96,14 @@ Remove (unlink) the FILE(s).\n\
   -v, --verbose         explain what is being done\n\
       --help            display this help and exit\n\
       --version         output version information and exit\n\
-"));
+\n\
+To remove a file whose name starts with a `-', for example `-foo',\n\
+use one of these commands:\n\
+  %s -- -foo\n\
+  %s ./-foo\n\
+"),
+	      program_name, program_name);
       puts (_("\nReport bugs to <bug-fileutils@gnu.org>."));
-      close_stdout ();
     }
   exit (status);
 }
@@ -127,6 +131,8 @@ main (int argc, char **argv)
   bindtextdomain (PACKAGE, LOCALEDIR);
   textdomain (PACKAGE);
 
+  atexit (close_stdout);
+
   rm_option_init (&x);
 
   while ((c = getopt_long (argc, argv, "dfirvR", long_opts, NULL)) != -1)
@@ -153,21 +159,13 @@ main (int argc, char **argv)
 	case 'v':
 	  x.verbose = 1;
 	  break;
+	case_GETOPT_HELP_CHAR;
+	case_GETOPT_VERSION_CHAR (PROGRAM_NAME, AUTHORS);
 	default:
 	  usage (1);
 	}
     }
 
-  if (show_version)
-    {
-      printf ("rm (%s) %s\n", GNU_PACKAGE, VERSION);
-      close_stdout ();
-      exit (0);
-    }
-
-  if (show_help)
-    usage (0);
-
   if (optind == argc)
     {
       if (x.ignore_missing_files)
@@ -198,7 +196,5 @@ main (int argc, char **argv)
 
   remove_fini ();
 
-  if (x.verbose)
-    close_stdout ();
   exit (fail);
 }
