@@ -83,6 +83,7 @@
 
 #include <getopt.h>
 #include <stdio.h>
+#include <assert.h>
 #include <setjmp.h>
 #include <signal.h>
 #include <sys/types.h>
@@ -116,7 +117,7 @@ char *xstrdup PARAMS ((char const *));
 # include <sys/time.h>	/* For struct timeval */
 # include <sys/stat.h>	/* For struct stat */
 
-# define GNU_PACKAGE "standalone"
+# define PACKAGE "standalone"
 # define VERSION "2.0" /* Kind of arbitrary... */
 
 # if __GNUC__ < 2 || __GNUC__ == 2 && __GNUC_MINOR__ < 5 || __STRICT_ANSI__
@@ -438,7 +439,7 @@ struct Options
 
 static struct option const long_opts[] =
 {
-  {"exact", required_argument, NULL, 'x'},
+  {"exact", no_argument, NULL, 'x'},
   {"force", no_argument, NULL, 'f'},
   {"iterations", required_argument, NULL, 'n'},
   {"size", required_argument, NULL, 's'},
@@ -507,16 +508,22 @@ files, most people use the --remove option.\n\
  * --------------------------------------------------------------------
  */
 
-#if ULONG_MAX == 0xffffffff
+#if defined __STDC__ && __STDC__
+# define UINT_MAX_32_BITS 4294967295U
+#else
+# define UINT_MAX_32_BITS 0xFFFFFFFF
+#endif
+
+#if ULONG_MAX == UINT_MAX_32_BITS
 typedef unsigned long word32;
 #else
-# if UINT_MAX == 0xffffffff
+# if UINT_MAX == UINT_MAX_32_BITS
 typedef unsigned word32;
 # else
-#  if USHRT_MAX == 0xffffffff
+#  if USHRT_MAX == UINT_MAX_32_BITS
 typedef unsigned short word32;
 #  else
-#   if UCHAR_MAX == 0xffffffff
+#   if UCHAR_MAX == UINT_MAX_32_BITS
 typedef unsigned char word32;
 #   else
      "No 32-bit type available!"
@@ -778,7 +785,7 @@ isaac_seed_finish (struct isaac_state *s)
 #define ISAAC_SEED(s,x) isaac_seed_data (s, &(x), sizeof (x))
 
 
-#if __GNUC__ >= 2 && (__i386__ || __alpha__ || _ARCH_PPC)
+#if __GNUC__ >= 2 && (__i386__ || __alpha__)
 /*
  * Many processors have very-high-resolution timer registers,
  * The timer registers can be made inaccessible, so we have to deal with the
@@ -814,6 +821,8 @@ isaac_seed_machdep (struct isaac_state *s)
       __asm__ __volatile__ ("rpcc %0" : "=r" (t));
 # endif
 # if _ARCH_PPC
+      /* Code not used because this instruction is available only on first-
+	 generation PPCs and evokes a SIGBUS on some Linux 2.4 kernels.  */
       word32 t;
       __asm__ __volatile__ ("mfspr %0,22" : "=r" (t));
 # endif
@@ -832,12 +841,12 @@ isaac_seed_machdep (struct isaac_state *s)
   }
 }
 
-#else /* !(__i386__ || __alpha__ || _ARCH_PPC) */
+#else /* !(__i386__ || __alpha__) */
 
 /* Do-nothing stub */
 # define isaac_seed_machdep(s) (void) 0
 
-#endif /* !(__i386__ || __alpha__ || _ARCH_PPC) */
+#endif /* !(__i386__ || __alpha__) */
 
 
 /*
@@ -990,13 +999,14 @@ fillpattern (int type, unsigned char *r, size_t size)
 }
 
 /*
- * Fill a buffer with random data.
- * size is rounded UP to a multiple of ISAAC_BYTES.
+ * Fill a buffer, R (of size SIZE_MAX), with random data.
+ * SIZE is rounded UP to a multiple of ISAAC_BYTES.
  */
 static void
-fillrand (struct isaac_state *s, word32 *r, size_t size)
+fillrand (struct isaac_state *s, word32 *r, size_t size_max, size_t size)
 {
   size = (size + ISAAC_BYTES - 1) / ISAAC_BYTES;
+  assert (size <= size_max);
 
   while (size--)
     {
@@ -1083,12 +1093,14 @@ dopass (int fd, char const *qname, off_t *sizep, int type,
       lim = sizeof r;
       if ((off_t) lim > size - offset && size != -1)
 	{
+	  if (size < offset)
+	    break;
 	  lim = (size_t) (size - offset);
 	  if (!lim)
 	    break;
 	}
       if (type < 0)
-	fillrand (s, r, lim);
+	fillrand (s, r, sizeof r, lim);
       /* Loop to retry partial writes. */
       for (soff = 0; soff < lim; soff += ssize)
 	{
@@ -1422,17 +1434,33 @@ do_wipefd (int fd, char const *qname, struct isaac_state *s,
   size = flags->size;
   if (size == -1)
     {
-      size = (S_ISREG (st.st_mode)
-	      ? st.st_size
-	      : lseek (fd, (off_t) 0, SEEK_END));
-      if (size < (S_ISREG (st.st_mode) ? 0 : -1))
+      /* Accept a length of zero only if it's a regular file.
+	 For any other type of file, try to get the size another way.  */
+      if (S_ISREG (st.st_mode))
 	{
-	  error (0, 0, _("%s: file has negative size"), qname);
-	  return -1;
+	  size = st.st_size;
+	  if (size < 0)
+	    {
+	      error (0, 0, _("%s: file has negative size"), qname);
+	      return -1;
+	    }
 	}
+      else
+	{
+	  size = lseek (fd, (off_t) 0, SEEK_END);
+	  if (size <= 0)
+	    {
+	      /* We are unable to determine the length, up front.
+		 Let dopass do that as part of its first iteration.  */
+	      size = -1;
+	    }
+	}
+
       if (0 <= size && !(flags->exact))
 	{
 	  size += ST_BLKSIZE (st) - 1 - (size - 1) % ST_BLKSIZE (st);
+
+	  /* If in rounding up, we've just overflowed, use the maximum.  */
 	  if (size < 0)
 	    size = TYPE_MAXIMUM (off_t);
 	}
